#include "ImageBuffer.h"
#include "Color.h"
#include <fstream>
#include <string>
#include <iostream>
#include "Vertex.h"
#include "Triangle.h"
#include "Line.h"
#include <vector>
#include "limits.h" 
#include "Matrix.h"

#define IMAGE_WIDTH 512
#define IMAGE_HEIGHT 512

using namespace std;

void LoadPartOne(vector<Triangle>& triangles, vector<Line>& lines, char* filename);
void RasterizeTriangle(const Triangle& tri, ImageBuffer& ib);
void RasterizeLine(const Line& line, ImageBuffer& ib);

int main(int argc, char* argv[])
{
	float left = -0.5;
	float right = 3.5;
	float bottom = -0.5;
	float top = 3.5;
	float near = -1.0;
	float far = 1.0;

	vector<Triangle> triangles;
	vector<Line> lines;
	LoadPartOne(triangles, lines, "example1.txt");
	ImageBuffer ib(IMAGE_WIDTH, IMAGE_HEIGHT);
	
	//viewport transformation - convert from cannonical to pixel
	Matrix vpt(IMAGE_WIDTH/2.0, 0.0, (IMAGE_WIDTH-1)/2.0, 
			0.0, IMAGE_HEIGHT/2.0, (IMAGE_HEIGHT-1)/2.0, 
			0.0, 0.0, 1.0);
	Matrix scale(2.0/(right-left), 0.0, 0.0, 0.0, 2.0/(top-bottom), 0.0, 0.0, 0.0, 1.0);
	Matrix trans(1.0, 0.0, (left+right)/-2.0, 0.0, 1.0, (bottom+top)/-2.0, 0.0, 0.0, 1.0);
	Matrix mat = vpt*scale*trans;
	
	for(int i=0; i<triangles.size(); i++)
	{
		//transfor vertices
		for(int j=0; j<3; j++)
		{
			triangles[i].v[j].pos = mat*triangles[i].v[j].pos;
		}
		//rasterize triangle
		RasterizeTriangle(triangles[i], ib);
	}

	for(int i=0; i<lines.size(); i++)
	{
		//transform vertices
		lines[i].v[0].pos = mat*lines[i].v[0].pos;
		lines[i].v[1].pos = mat*lines[i].v[1].pos;
		//rasterize line
		RasterizeLine(lines[i], ib);
	}

	ib.writeBMP("test.bmp");
	//char c;
	//cin >>  c;
	return 0;
}

void LoadPartOne( vector<Triangle>& triangles, vector<Line>& lines, char* filename)
{
	string line;
	ifstream file(filename);
	if(file.is_open())
	{
		int numTriangles = 0;
		file >> numTriangles;
		//cout << "numTriangles=" << numTriangles << endl;
		int x = 0; 
		int y = 0;
		int r = 0;
		int g = 0; 
		int b = 0;
		int a = 0;
		for(int i=0; i<numTriangles; i++)
		{	
			Vertex v[3];
			for(int j=0; j<3; j++)
			{
				file >> x;
				file >> y;
				file >> r;
				file >> g;
				file >> b;
				file >> a;
				//cout << "trivert=" << x << "," << y << "," << r << "," << g << "," << b << "," << a << endl;
				v[j] = Vertex(x,y,1.0,Color(r,g,b,a)); //need to avoid duplication here eventually ??
			}
			triangles.push_back(Triangle(v[0], v[1], v[2]));
		}
		int numLines = 0;
		file >> numLines;
		//cout << "numLines=" << numLines;
		Vertex v[2];
		for(int i=0; i<numLines; i++)
		{
			for(int j=0; j<2; j++)
			{
				file >> x;
				file >> y;
				file >> r;
				file >> g;
				file >> b;
				file >> a;
				//cout << "linevert=" << x << "," << y << "," << r << "," << g << "," << b << "," << a << endl;
				v[j] = Vertex(x,y,1.0,Color(r,g,b,a));
			}
			lines.push_back(Line(v[0], v[1]));
		}
	}
	else
	{
		cout << "Couldn't open file";
	}
}

void RasterizeTriangle(const Triangle& tri, ImageBuffer& ib)
{
	int xmin = INT_MAX; 
	int xmax = INT_MIN;
	int ymin = INT_MAX;
	int ymax = INT_MIN;
	for(int i=0; i<3; i++)
	{
		if(tri.v[i].pos.e[0] < xmin)
		{
			xmin = tri.v[i].pos.e[0];
		}
		if(tri.v[i].pos.e[0] > xmax)
		{
			xmax = tri.v[i].pos.e[0];
		}
		if(tri.v[i].pos.e[1] < ymin)
		{
			ymin = tri.v[i].pos.e[1];  //do we need floor celining.  should we preserve floats?
		}
		if(tri.v[i].pos.e[1] > ymax)
		{
			ymax = tri.v[i].pos.e[1];
		}
	}

	//cout << "ymin,max=" << ymin << ", " << ymax << endl;
	//cout << "xmin,max=" << xmin << ", " << xmax << endl;

	float x0 = tri.v[0].pos.e[0];
	float x1 = tri.v[1].pos.e[0];
	float x2 = tri.v[2].pos.e[0];
	float y0 = tri.v[0].pos.e[1];
	float y1 = tri.v[1].pos.e[1];
	float y2 = tri.v[2].pos.e[1];

	float x01 = y0-y1;
	float x12 = y1-y2;
	float x20 = y2-y0;
	float y01 = x1-x0;
	float y12 = x2-x1;
	float y20 = x0-x2;
	float c01 = (x0*y1) - (x1*y0);
	float c12 = (x1*y2) - (x2*y1);
	float c20 = (x2*y0) - (x0*y2);

	float vertf01 = (x01*x2) + (y01*y2) + c01;
	float vertf12 = (x12*x0) + (y12*y0) + c12;
	float vertf20 = (x20*x1) + (y20*y1) + c20;

	for(int y=ymin; y<ymax; y++)
	{
		for(int x=xmin; x<xmax; x++)
		{
			float f01 = (x01*x) + (y01*y) + c01;
			float f12 = (x12*x) + (y12*y) + c12;
			float f20 = (x20*x) + (y20*y) + c20;
			float alpha = f12/vertf12;
			float beta = f20/vertf20;
			float gamma = f01/vertf01; 
			if(alpha > 0 && beta > 0 && gamma > 0)
			{
				
				ib.set(x, y, (tri.v[0].color*alpha) + (tri.v[1].color*beta) + (tri.v[2].color*gamma));
			}
		}
	}
}

void RasterizeLine(const Line& line, ImageBuffer& ib)
{
	float y0 = line.v[0].pos.e[1];
	float y1 = line.v[1].pos.e[1];
	float x0 = line.v[0].pos.e[0];
	float x1 = line.v[1].pos.e[0];
	float r0 = line.v[0].color.r;
	float g0 = line.v[0].color.g;
	float b0 = line.v[0].color.b;
	float r1 = line.v[1].color.r;
	float g1 = line.v[1].color.g;
	float b1 = line.v[1].color.b;

	if(x0 > x1)
	{
		//swap 0 and 1  so we can loop from x0 to x1
		float xtemp = x0;
		x0 = x1;
		x1 = xtemp;
		float ytemp = y0;
		y0 = y1;
		y1 = ytemp;
		float rtemp = r0;
		r0 = r1;
		r1 = rtemp;
		float gtemp = g0;
		g0 = g1;
		g1 = gtemp;
		float btemp = b0;
		b0 = b1;
		b1 = btemp;
	}

	float invx1subx0 = 1.0/(x1 - x0);
	float dy = (y1 - y0) * invx1subx0;
	float dr = (r1 - r0) * invx1subx0;
	float dg = (g1 - g0) * invx1subx0;
	float db = (b1 - b0) * invx1subx0;

	//cout << "line " << x0  << " " <<  x1 << endl;	

	float y = y0;
	float r = r0;
	float g = g0;
	float b = b0;

	//cout << "start" << endl;

	for(int x= (int)x0; x < x1; x++)
	{
		int yround = (int) y + 0.5;
		ib.set(x, yround, Color(r,g,b,1.0));
		//cout << "x=" << x << " y=" << yround << endl; 
		y = y + dy;
		r = r + dr;
		g = g + dg;
		b = b + db;
	}
}

/*void RasterizeLine(const Line& line, ImageBuffer& ib)
{
	float y0 = line.v[0].pos.e[1];
	float y1 = line.v[1].pos.e[1];
	float x0 = line.v[0].pos.e[0];
	float x1 = line.v[1].pos.e[0];
	float y = (int)y0;
	float d = (2.0*(y0-y1)*(x1+1.0)) + (x1 - x0)*((2.0*y0) + 1.0) + (2.0*x0*y1) - (2.0*x1*y0);
	for(int x = (int)x0; x < (int)x1; x++)
	{
		ib.set(x, y, Color(1,0,0,1));
		if(d<0.0)
		{
			y = y+1.0;
			d = d + (2.0*(x1-x0)) + (2.0*(y0-y1));
		}
		else
		{
			d = d + (2.0*(y0-y1));
		}
	}
}*/